import os
import time
import numpy as np
from models import *
import matplotlib.pyplot as plt
import torch
import torch.optim
from utils.denoising_utils import *

torch.set_default_tensor_type("torch.cuda.FloatTensor")
plt.rcParams.update({"font.size": 18, "font.family": "serif"})


def compare_psnr(x, gt):
    out = 10 * np.log10(1 / ((x - gt) ** 2).mean().item())
    return out


def denoise_image(img_np, sigma):
    sigma_ = sigma / 255.0
    img_pil = np_to_pil(img_np)
    img_noisy_pil, img_noisy_np = get_noisy_image(img_np, sigma_)

    INPUT = "noise"
    pad = "reflection"
    OPT_OVER = "net"
    reg_noise_std = 1.0 / 30.0
    LR = 0.01
    OPTIMIZER = "adam"
    show_every = 10
    exp_weight = 0.99
    num_iter = 3000

    input_depth = 32

    net = get_net(input_depth, "skip", pad, skip_n33d=128, skip_n33u=128, skip_n11=4, num_scales=5, upsample_mode="bilinear").cuda()

    net_input = get_noise(input_depth, INPUT, (img_pil.size[1], img_pil.size[0])).detach()

    mse = torch.nn.MSELoss()
    img_noisy_torch = np_to_torch(img_noisy_np)
    net_input_saved = net_input.detach().clone()
    noise = net_input.detach().clone()
    global i, out_avg, psnr_noisy_last, last_net
    out_avg = None
    last_net = None
    psnr_noisy_last = 0
    i = 0

    def closure():
        global i, out_avg, psnr_noisy_last, last_net, net_input
        if reg_noise_std > 0:
            net_input = net_input_saved + (noise.normal_() * reg_noise_std)
        out = net(net_input)

        if out_avg is None:
            out_avg = out.detach()
        else:
            out_avg = out_avg * exp_weight + out.detach() * (1 - exp_weight)

        total_loss = mse(out, img_noisy_torch.cuda())
        total_loss.backward()

        psnr_noisy = compare_psnr(img_noisy_np, img_np)
        psnr_gt = compare_psnr(img_np, out.detach().cpu().numpy()[0])
        psnr_gt_sm = compare_psnr(img_np, out_avg.detach().cpu().numpy()[0])

        if i % show_every == 0:
            print(
                "Iteration %05d Loss %f PSNR_noisy: %f psnr_gt: %f PSNR_gt_sm: %f" % (i, total_loss.item(), psnr_noisy, psnr_gt, psnr_gt_sm),
                "\n",
                end="",
            )

        if i % show_every:
            if psnr_noisy - psnr_noisy_last < -5:
                print("Falling back to previous checkpoint.")
                for new_param, net_param in zip(last_net, net.parameters()):
                    net_param.data.copy_(new_param.cuda())
                return total_loss * 0
            else:
                last_net = [x.detach().cpu() for x in net.parameters()]
                psnr_noisy_last = psnr_noisy

        i += 1
        return total_loss

    p = get_params(OPT_OVER, net, net_input)
    optimize(OPTIMIZER, p, closure, LR, num_iter)

    out_np = torch_to_np(net(net_input))

    return out_np, img_noisy_np


out_avg = None
last_net = None
psnr_noisy_last = 0
i = 0

# Load BSDS-300 dataset
dataset_path = "./BSDS300_subset/"
selected_images = []
for filename in os.listdir(dataset_path):
    if filename.endswith(".jpg") or filename.endswith(".png"):
        img_path = os.path.join(dataset_path, filename)
        selected_images.append(crop_image(get_image(img_path, -1)[0], d=32))

# Define noise levels
noise_levels = [15, 25, 50]

# Denoise images and calculate PSNR for each noise level
for sigma in noise_levels:
    psnr_values = []
    for img_pil in selected_images:
        print(f"Processing image {selected_images.index(img_pil) + 1}/{len(selected_images)}")
        start = time.time()
        img_np = pil_to_np(img_pil)
        denoised_np, noisy_np = denoise_image(img_np, sigma)
        psnr_values.append(compare_psnr(denoised_np, img_np))
        end = time.time()

        # Save the noised and denoised images for the first image in the selected set
        if img_pil == selected_images[0]:
            print(f"Time: {end - start:.2f}s")

            plt.figure(figsize=(18, 6))
            plt.subplot(1, 3, 1)
            plt.imshow(img_np.transpose(1, 2, 0))
            plt.title("Original")
            plt.axis("off")

            plt.subplot(1, 3, 2)
            plt.imshow(noisy_np.transpose(1, 2, 0))
            plt.title(f"Noisy (sigma = {sigma})")
            plt.axis("off")

            plt.subplot(1, 3, 3)
            plt.imshow(np.clip(denoised_np, 0, 1).transpose(1, 2, 0))
            plt.title(f"Denoised (sigma = {sigma})")
            plt.axis("off")

            plt.tight_layout()
            plt.savefig(f"deep_prior_sigma_{sigma}.png", bbox_inches="tight", dpi=200)
            plt.close()

    print(f"Noise Level: {sigma}, Average PSNR: {np.mean(psnr_values):.2f}")
