import os
import time
import numpy as np
import matplotlib.pyplot as plt
import torch
import torch.optim
from utils.denoising_utils import *
import tinycudann as tcnn
import commentjson as json

plt.rcParams.update({"font.size": 18, "font.family": "serif"})


def compare_psnr(x, gt):
    out = 10 * np.log10(1 / ((x - gt) ** 2).mean().item())
    return out


def fit_image(img_np, sigma, config):
    sigma_ = sigma / 255.0
    img_pil = np_to_pil(img_np)
    img_noisy_pil, img_noisy_np = get_noisy_image(img_np, sigma_)

    device = torch.device("cuda")
    resolution = img_np.shape[1:]
    n_channels = img_np.shape[0]

    encoding = tcnn.Encoding(n_input_dims=2, encoding_config=config["encoding"], dtype=torch.float32)
    network = torch.nn.Sequential(
        torch.nn.Linear(encoding.n_output_dims, 128, bias=False),
        torch.nn.LeakyReLU(0.1),
        torch.nn.Linear(128, 64, bias=False),
        torch.nn.LeakyReLU(0.1),
        torch.nn.Linear(64, 3, bias=False),
    ).to(device)
    model = torch.nn.Sequential(encoding, network)

    optimizer = torch.optim.Adam(model.parameters(), lr=0.01)

    n_pixels = resolution[0] * resolution[1]

    half_dx = 0.5 / resolution[0]
    half_dy = 0.5 / resolution[1]
    xs = torch.linspace(half_dx, 1 - half_dx, resolution[0], device=device)
    ys = torch.linspace(half_dy, 1 - half_dy, resolution[1], device=device)
    xv, yv = torch.meshgrid([xs, ys])

    xy = torch.stack((yv.flatten(), xv.flatten())).t()

    img_noisy_torch = np_to_torch(img_noisy_np).to(device)
    img_torch = np_to_torch(img_np).to(device)

    num_iter = 1000
    show_every = 10
    exp_weight = 0.99
    out_avg = None
    psnr_noisy_last = 0
    i = 0

    for i in range(num_iter):
        batch = xy
        targets = img_noisy_torch.squeeze().permute(1, 2, 0).view(n_pixels, n_channels)
        out = model(batch)

        if out_avg is None:
            out_avg = out.detach()
        else:
            out_avg = out_avg * exp_weight + out.detach() * (1 - exp_weight)

        total_loss = torch.nn.functional.mse_loss(out, targets)

        optimizer.zero_grad()
        total_loss.backward()
        optimizer.step()

        psnr_noisy = compare_psnr(img_noisy_np, img_np)
        psnr_gt = compare_psnr(img_np, out.view(*resolution, n_channels).permute(2, 0, 1).detach().cpu().numpy())
        psnr_gt_sm = compare_psnr(img_np, out_avg.view(*resolution, n_channels).permute(2, 0, 1).detach().cpu().numpy())

        if i % show_every == 0:
            print(
                "Iteration %05d Loss %f PSNR_noisy: %f psnr_gt: %f PSNR_gt_sm: %f" % (i, total_loss.item(), psnr_noisy, psnr_gt, psnr_gt_sm),
                "\n",
                end="",
                flush=True,
            )

        if i % show_every:
            if psnr_noisy - psnr_noisy_last < -5:
                print("Falling back to previous checkpoint.")
                for new_param, net_param in zip(last_net, model.parameters()):
                    net_param.data.copy_(new_param.cuda())
                break
            else:
                last_net = [x.detach().cpu() for x in model.parameters()]
                psnr_noisy_last = psnr_noisy

    out_np = out_avg.view(*resolution, n_channels).permute(2, 0, 1).detach().cpu().numpy()
    return out_np, img_noisy_np


# Load BSDS-300 dataset
dataset_path = "./BSDS300_subset/"
selected_images = []
for filename in os.listdir(dataset_path):
    if filename.endswith(".jpg") or filename.endswith(".png"):
        img_path = os.path.join(dataset_path, filename)
        selected_images.append(crop_image(get_image(img_path, -1)[0], d=32))


# Load tiny-cuda-nn config
# with open("./config.json") as config_file:
# with open("./config_freq.json") as config_file:
with open("./config_none.json") as config_file:
    config = json.load(config_file)

# Define noise levels
noise_levels = [15, 25, 50]

# Fit images and calculate PSNR for each noise level
for sigma in noise_levels:
    psnr_values = []
    for img_pil in selected_images:
        print(f"Fitting image number {selected_images.index(img_pil) + 1} for noise level {sigma}...")
        start = time.time()
        img_np = pil_to_np(img_pil)
        fitted_np, noisy_np = fit_image(img_np, sigma, config)
        psnr_values.append(compare_psnr(fitted_np, img_np))
        end = time.time()

        # Save the noised and fitted images for the first image in the selected set
        if img_pil == selected_images[0]:
            print(f"Image fitting time: {end - start:.2f}s")

            plt.figure(figsize=(18, 6))
            plt.subplot(1, 3, 1)
            plt.imshow(img_np.transpose(1, 2, 0))
            plt.title("Original")
            plt.axis("off")

            plt.subplot(1, 3, 2)
            plt.imshow(noisy_np.transpose(1, 2, 0))
            plt.title(f"Noisy (sigma = {sigma})")
            plt.axis("off")

            plt.subplot(1, 3, 3)
            plt.imshow(np.clip(fitted_np, 0, 1).transpose(1, 2, 0))
            plt.title(f"Denoised (sigma = {sigma})")
            plt.axis("off")

            plt.tight_layout()
            plt.savefig(f"tcnn_sigma_{sigma}_none.png", bbox_inches="tight", dpi=200)
            plt.close()

    print(f"Noise Level: {sigma}, Average PSNR: {np.mean(psnr_values):.2f}")
